# -*- coding: utf-8 -*-
from plone.testing import z2
from plone.testing import zca

import plone.autoform


AUTOFORM_FIXTURE = zca.ZCMLSandbox(
    bases=(z2.STARTUP,),
    filename='configure.zcml',
    package=plone.autoform,
    name='plone.autoform:Fixture')

AUTOFORM_INTEGRATION_TESTING = z2.IntegrationTesting(
    bases=(AUTOFORM_FIXTURE,),
    name='plone.autoform:Integration')



# -*- coding: utf-8 -*-
from plone.testing import layered
from plone.testing.zca import UNIT_TESTING

import doctest
import unittest


def test_suite():
    return unittest.TestSuite((
        layered(
            doctest.DocFileSuite(
                '../autoform.rst',
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,),
            layer=UNIT_TESTING),
        layered(
            doctest.DocFileSuite(
                'subform.txt',
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,),
            layer=UNIT_TESTING),
        layered(
            doctest.DocFileSuite(
                '../view.txt',
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,),
            layer=UNIT_TESTING),
        layered(
            doctest.DocFileSuite(
                '../supermodel.txt',
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,),
            layer=UNIT_TESTING),
    ))



# -*- coding: utf-8 -*-
from plone.testing.zca import UNIT_TESTING

import unittest


class TestParameterizedWidget(unittest.TestCase):

    layer = UNIT_TESTING

    def test_widget_instantiated_with_parameters(self):
        from plone.autoform.widgets import ParameterizedWidget
        from z3c.form.interfaces import IFieldWidget
        from z3c.form.interfaces import IWidget
        from zope.interface import implementer
        from zope.schema import Field

        @implementer(IWidget)
        class DummyWidget(object):

            def __init__(self, request):
                self.request = request

        @implementer(IFieldWidget)
        def DummyFieldWidget(field, request):
            return DummyWidget(request)

        field = Field()
        request = object()
        widget = ParameterizedWidget(DummyWidget, foo='bar')(field, request)

        self.assertTrue(isinstance(widget, DummyWidget))
        self.assertEqual('bar', widget.foo)

    def test_default_widget_instantiated(self):
        from plone.autoform.widgets import ParameterizedWidget
        from z3c.form.interfaces import IFieldWidget
        from zope.component import provideAdapter
        from zope.interface import Interface
        from zope.interface import implementer
        from zope.schema import Field

        class DummyWidget(object):

            def __init__(self, request):
                self.request = request

        @implementer(IFieldWidget)
        def DummyFieldWidget(field, request):
            return DummyWidget(request)

        provideAdapter(DummyFieldWidget, (Interface, Interface), IFieldWidget)

        field = Field()
        request = object()
        widget = ParameterizedWidget(foo='bar')(field, request)

        self.assertTrue(isinstance(widget, DummyWidget))
        self.assertEqual('bar', widget.foo)

    def test_validates_for_field_widget(self):
        from plone.autoform.widgets import ParameterizedWidget

        class NotAWidget(object):
            pass

        try:
            x = ParameterizedWidget(NotAWidget)
            x  # PEP8
        except TypeError:
            pass
        else:
            self.fail('Expected TypeError')



# -*- coding: utf-8 -*-
from plone.autoform import directives as form
from plone.autoform.interfaces import MODES_KEY
from plone.autoform.interfaces import OMITTED_KEY
from plone.autoform.interfaces import ORDER_KEY
from plone.autoform.interfaces import READ_PERMISSIONS_KEY
from plone.autoform.interfaces import WIDGETS_KEY
from plone.autoform.interfaces import WRITE_PERMISSIONS_KEY
from plone.autoform.testing import AUTOFORM_INTEGRATION_TESTING
from plone.supermodel import model
from zope.interface import Interface

import unittest
import zope.schema


class DummyWidget(object):
    pass


class TestSchemaDirectives(unittest.TestCase):

    layer = AUTOFORM_INTEGRATION_TESTING

    def test_schema_directives_store_tagged_values(self):

        class IDummy(model.Schema):

            form.omitted('foo', 'bar')
            form.omitted(model.Schema, 'qux')
            form.no_omit(model.Schema, 'bar')
            form.widget(foo='some.dummy.Widget', baz='other.Widget')
            form.mode(bar='hidden')
            form.mode(model.Schema, bar='input')
            form.order_before(baz='title')
            form.order_after(qux='title')
            form.read_permission(foo='zope2.View')
            form.write_permission(foo='cmf.ModifyPortalContent')

            foo = zope.schema.TextLine(title=u'Foo')
            bar = zope.schema.TextLine(title=u'Bar')
            baz = zope.schema.TextLine(title=u'Baz')
            qux = zope.schema.TextLine(title=u'Qux')

        model.finalizeSchemas(IDummy)

        self.assertEqual({'foo': 'some.dummy.Widget',
                          'baz': 'other.Widget'},
                         IDummy.queryTaggedValue(WIDGETS_KEY))
        self.assertEqual([(Interface, 'foo', 'true'),
                          (Interface, 'bar', 'true'),
                          (model.Schema, 'qux', 'true'),
                          (model.Schema, 'bar', 'false')],
                         IDummy.queryTaggedValue(OMITTED_KEY))
        self.assertEqual([(Interface, 'bar', 'hidden'),
                          (model.Schema, 'bar', 'input')],
                         IDummy.queryTaggedValue(MODES_KEY))
        self.assertEqual([('baz', 'before', 'title',),
                          ('qux', 'after', 'title')],
                         IDummy.queryTaggedValue(ORDER_KEY))
        self.assertEqual({'foo': 'zope2.View'},
                         IDummy.queryTaggedValue(READ_PERMISSIONS_KEY))
        self.assertEqual({'foo': 'cmf.ModifyPortalContent'},
                         IDummy.queryTaggedValue(WRITE_PERMISSIONS_KEY))

    def test_widget_supports_instances_and_strings(self):

        class IDummy(model.Schema):
            form.widget(foo=DummyWidget)

            foo = zope.schema.TextLine(title=u'Foo')
            bar = zope.schema.TextLine(title=u'Bar')
            baz = zope.schema.TextLine(title=u'Baz')

        self.assertEqual(
            {'foo': 'plone.autoform.tests.test_directives.DummyWidget'},
            IDummy.queryTaggedValue(WIDGETS_KEY)
        )

    def test_widget_parameterized(self):
        from zope.interface import implementer
        from z3c.form.interfaces import IWidget
        from plone.autoform.widgets import ParameterizedWidget

        @implementer(IWidget)
        class DummyWidget(object):

            def __init__(self, request):
                pass

        class IDummy(model.Schema):
            form.widget('foo', DummyWidget, foo='bar')
            foo = zope.schema.TextLine(title=u'Foo')

        tv = IDummy.queryTaggedValue(WIDGETS_KEY)
        self.assertTrue(isinstance(tv['foo'], ParameterizedWidget))
        self.assertTrue(tv['foo'].widget_factory is DummyWidget)
        self.assertEqual('bar', tv['foo'].params['foo'])

    def test_widget_parameterized_default_widget_factory(self):
        from zope.interface import implementer
        from z3c.form.interfaces import IWidget
        from plone.autoform.widgets import ParameterizedWidget

        @implementer(IWidget)
        class DummyWidget(object):

            def __init__(self, request):
                pass

        class IDummy(model.Schema):
            form.widget('foo', foo='bar')
            foo = zope.schema.TextLine(title=u'Foo')

        tv = IDummy.queryTaggedValue(WIDGETS_KEY)
        self.assertTrue(isinstance(tv['foo'], ParameterizedWidget))
        self.assertTrue(tv['foo'].widget_factory is None)
        self.assertEqual('bar', tv['foo'].params['foo'])

    def test_widget_parameterized_wrong_type(self):
        try:
            class IDummy(model.Schema):
                form.widget('foo', object())
        except TypeError:
            pass
        else:
            self.fail('Expected TypeError')

    def test_multiple_invocations(self):

        class IDummy(model.Schema):

            form.omitted('foo')
            form.omitted('bar')
            form.widget(foo='some.dummy.Widget')
            form.widget(baz='other.Widget')
            form.mode(bar='hidden')
            form.mode(foo='display')
            form.order_before(baz='title')
            form.order_after(baz='qux')
            form.order_after(qux='bar')
            form.order_before(foo='body')
            form.read_permission(foo='zope2.View', bar='zope2.View')
            form.read_permission(baz='random.Permission')
            form.write_permission(foo='cmf.ModifyPortalContent')
            form.write_permission(baz='another.Permission')

            foo = zope.schema.TextLine(title=u'Foo')
            bar = zope.schema.TextLine(title=u'Bar')
            baz = zope.schema.TextLine(title=u'Baz')
            qux = zope.schema.TextLine(title=u'Qux')

        self.assertEqual({'foo': 'some.dummy.Widget',
                          'baz': 'other.Widget'},
                         IDummy.queryTaggedValue(WIDGETS_KEY))
        self.assertEqual([(Interface, 'foo', 'true'),
                          (Interface, 'bar', 'true')],
                         IDummy.queryTaggedValue(OMITTED_KEY))
        self.assertEqual([(Interface, 'bar', 'hidden'),
                          (Interface, 'foo', 'display')],
                         IDummy.queryTaggedValue(MODES_KEY))
        self.assertEqual([('baz', 'before', 'title'),
                          ('baz', 'after', 'qux'),
                          ('qux', 'after', 'bar'),
                          ('foo', 'before', 'body'), ],
                         IDummy.queryTaggedValue(ORDER_KEY))
        self.assertEqual(
            {'foo': 'zope2.View',
             'bar': 'zope2.View',
             'baz': 'random.Permission'},
            IDummy.queryTaggedValue(READ_PERMISSIONS_KEY)
        )
        self.assertEqual(
            {'foo': 'cmf.ModifyPortalContent', 'baz': 'another.Permission'},
            IDummy.queryTaggedValue(WRITE_PERMISSIONS_KEY)
        )

    def test_misspelled_field(self):

        try:
            class IBar(model.Schema):
                form.order_before(ber='*')
                bar = zope.schema.TextLine()
        except ValueError:
            pass
        else:
            self.fail('Did not raise ValueError')

        try:
            class IBaz(model.Schema):
                form.omitted('buz')
                baz = zope.schema.TextLine()
        except ValueError:
            pass
        else:
            self.fail('Did not raise ValueError')

    def test_derived_class_fields(self):

        class IFoo(model.Schema):
            foo = zope.schema.TextLine()

        class IBar(IFoo):
            form.order_after(foo='bar')
            bar = zope.schema.TextLine()

        self.assertEqual(
            [('foo', 'after', 'bar'), ],
            IBar.queryTaggedValue(ORDER_KEY)
        )


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(TestSchemaDirectives),
    ))



# -*- coding: utf-8 -*-
from collections import OrderedDict
from plone.testing.zca import UNIT_TESTING

import unittest


class TestBase(unittest.TestCase):

    layer = UNIT_TESTING

    def test_calc_field_move_basics(self):
        from plone.autoform.base import AutoFields
        autofields = AutoFields()

        # we have a schema with fields [a, b, c]
        # field 'c' after 'a'
        order = [('c', 'after', 'a'), ]
        rules = autofields._calculate_field_moves(order)
        self.assertIn('a', rules)
        self.assertNotIn('c', rules)
        self.assertIn('with', rules['a'])
        self.assertIn('stub', rules['a'])
        self.assertIn('c', rules['a']['with'])
        self.assertIn('parent', rules['a']['with']['c'])
        self.assertIs(rules['a'], rules['a']['with']['c']['parent'])
        self.assertIn('target', rules['a']['with']['c'])
        self.assertIn('dir', rules['a']['with']['c'])

    def test_calc_field_move_simple_dependency(self):
        from plone.autoform.base import AutoFields
        autofields = AutoFields()

        # we have a schema with fields [a, b, c]
        # field a after b and field 'c' after 'a'
        order = [
            ('a', 'after', 'b'),
            ('c', 'after', 'a'),
        ]
        expected = {
            'stub': True,
            'with': OrderedDict(
                [
                    (
                        'a',
                        {
                            'with': OrderedDict(
                                [('c', {'target': 'a', 'dir': 'after'})]
                            ),
                            'target': 'b', 'dir': 'after'}
                    ),
                ]
            )
        }
        rules = autofields._calculate_field_moves(order)
        self.assertIn('b', rules)
        self.assertNotIn('a', rules)
        self.assertNotIn('c', rules)

        # remove parent key enable comparison of  dicts
        del rules['__all__']['a']['parent']
        del rules['__all__']['c']['parent']
        self.assertEqual(rules['b'], expected)

        # we have a schema with fields [a, b, c]
        # vice versa defined now, must lead to same result
        # field 'c' after 'a' and field a after b
        order = reversed(order)
        rules = autofields._calculate_field_moves(order)
        self.assertIn('b', rules)
        self.assertNotIn('a', rules)
        self.assertNotIn('c', rules)

        # remove parent key enable comparison of  dicts
        del rules['__all__']['a']['parent']
        del rules['__all__']['c']['parent']
        self.assertEqual(rules['b'], expected)

    def test_calc_field_move_multiple_dependencies(self):
        from plone.autoform.base import AutoFields
        autofields = AutoFields()

        order = [
            ('a', 'after', 'b'),
            ('c', 'after', 'a'),
            ('d', 'after', 'c'),
            ('z', 'after', 'x'),
            ('x', 'after', 'y'),
        ]
        rules = autofields._calculate_field_moves(order)
        self.assertIn('b', rules)
        self.assertIn('y', rules)
        self.assertNotIn('a', rules)
        self.assertNotIn('c', rules)
        self.assertNotIn('d', rules)
        self.assertNotIn('z', rules)
        self.assertNotIn('x', rules)

        self.assertIn('a', rules['b']['with'])
        self.assertEqual(1, len(rules['b']['with']))

        self.assertIn('c', rules['b']['with']['a']['with'])
        self.assertEqual(1, len(rules['b']['with']['a']['with']))

        self.assertIn('d', rules['b']['with']['a']['with']['c']['with'])
        self.assertEqual(1, len(rules['b']['with']['a']['with']['c']['with']))

    def test_calc_field_move_override(self):
        from plone.autoform.base import AutoFields
        autofields = AutoFields()

        order = [
            ('c', 'after', 'a'),
            ('a', 'after', 'b'),
            ('c', 'after', 'z'),
        ]
        rules = autofields._calculate_field_moves(order)
        self.assertIn('b', rules)
        self.assertIn('z', rules)
        self.assertNotIn('a', rules)
        self.assertNotIn('c', rules)

    def test_updateFieldsFromSchemata(self):
        from plone.autoform.base import AutoFields
        autofields = AutoFields()
        autofields.request = {}
        autofields.updateFieldsFromSchemata()
        self.assertEqual(autofields.groups, [])



# -*- coding: utf-8 -*-
from collections import OrderedDict
from operator import attrgetter
from plone.autoform.interfaces import ORDER_KEY
from plone.autoform.utils import processFields
from plone.supermodel.interfaces import DEFAULT_ORDER
from plone.supermodel.utils import mergedTaggedValueList
from plone.z3cform.fieldsets.group import GroupFactory
from plone.z3cform.fieldsets.utils import move
from z3c.form import field
from z3c.form.util import expandPrefix

import logging


logger = logging.getLogger(__name__)
_marker = object()


class AutoFields(object):
    """Mixin class for the WidgetsView and AutoExtensibleForm classes.
    Takes care of actually processing field updates
    """

    schema = None
    additionalSchemata = ()

    fields = field.Fields()
    groups = ()

    ignorePrefix = False
    autoGroups = False

    def updateFieldsFromSchemata(self):

        # If the form is called from the ++widget++ traversal namespace,
        # we won't have a user yet. In this case, we can't perform permission
        # checks.

        have_user = bool(self.request.get('AUTHENTICATED_USER', False))

        # Turn fields into an instance variable, since we will be modifying it
        self.fields = field.Fields(self.fields)

        # Copy groups to an instance variable and ensure that we have
        # the more mutable factories, rather than 'Group' subclasses

        groups = []

        for group in self.groups:
            group_name = getattr(group, '__name__', group.label)
            fieldset_group = GroupFactory(
                group_name,
                field.Fields(group.fields),
                group.label,
                getattr(group, 'description', None)
            )
            groups.append(fieldset_group)

        # Copy to instance variable only after we have potentially read from
        # the class
        self.groups = groups

        prefixes = {}

        if self.schema is not None:
            processFields(self, self.schema, permissionChecks=have_user)

        # Set up all widgets, modes, omitted fields and fieldsets
        for schema in self.additionalSchemata:

            # Find the prefix to use for this form and cache for next round
            prefix = self.getPrefix(schema)
            if prefix and prefix in prefixes:
                prefix = schema.__identifier__
            prefixes[schema] = prefix

            # By default, there's no default group, i.e. fields go
            # straight into the default fieldset

            defaultGroup = None

            # Create groups from schemata if requested and set default
            # group

            if self.autoGroups:
                # use interface name, or prefix for anonymous schema
                group_name = schema.__name__ or prefix or None

                # Look for group - note that previous processFields
                # may have changed the groups list, so we can't easily
                # store this in a dict.
                found = False
                for g in self.groups:
                    if group_name == getattr(g, '__name__', g.label):
                        found = True
                        break

                if not found:
                    fieldset_group = GroupFactory(
                        group_name,
                        field.Fields(),
                        label=group_name,
                        description=schema.__doc__,
                        order=DEFAULT_ORDER,
                    )
                    self.groups.append(fieldset_group)

                defaultGroup = group_name

            processFields(
                self,
                schema,
                prefix=prefix,
                defaultGroup=defaultGroup,
                permissionChecks=have_user
            )

        # Then process relative field movements. The base schema is processed
        # last to allow it to override any movements made in additional
        # schemata.
        rules = {'__all__': {}}
        for schema in self.additionalSchemata:
            order = mergedTaggedValueList(schema, ORDER_KEY)
            rules = self._calculate_field_moves(
                order,
                prefix=prefixes[schema],
                rules=rules,
            )
        if self.schema is not None:
            order = mergedTaggedValueList(self.schema, ORDER_KEY)
            rules = self._calculate_field_moves(order, rules=rules)
        self._cleanup_rules(rules)
        self._process_field_moves(rules)
        self._process_group_order()

    def getPrefix(self, schema):
        """Get the preferred prefix for the given schema
        """
        if self.ignorePrefix:
            return ''
        return schema.__name__

    def _prepare_names(self, source, target, prefix):
            # calculate prefixed fieldname
            if prefix:
                source = '{0}.{1}'.format(prefix, source)

            # Handle shortcut: leading . means "in this form". May be useful
            # if you want to move a field relative to one in the current
            # schema or (more likely) a base schema of the current schema,
            # without having to repeat the full prefix of this schema.
            if target.startswith('.'):
                target = target[1:]
                if prefix:
                    target = expandPrefix(prefix) + target
            return source, target

    def _cleanup_rules(self, rules):
        for rulename in rules['__all__']:
            if 'parent' in rules['__all__'][rulename]:
                del rules['__all__'][rulename]['parent']
        del rules['__all__']

    def _calculate_field_moves(self, order, prefix='', rules=None):
        """Calculates all needed field rules
        """
        # we want to be independent from the order of the schemas coming later
        # so a if field_c is first moved after field_a, then field_a is moved
        # after field_c, the output should be: b, a, c, because or first move
        # sticks
        if rules is None:
            rules = {}
        allrules = rules.get('__all__', None)
        if allrules is None:
            allrules = rules['__all__'] = dict()

        # (current field name, 'before'/'after', other field name)
        for source, direction, target in order:
            source, target = self._prepare_names(source, target, prefix)
            # use a simple tree to resolve dependencies
            rule = allrules.get(source, {})
            if (
                'target' in rule and target != rule['target']
            ):
                # target override
                # reset this rule to a stub first
                del rule['target']
                del rule['dir']
                rule['stub'] = True
                # unlink in parent
                del rule['parent']['with'][source]
                del rule['parent']
            if (
                'dir' in rule and direction != rule['dir']
            ):
                # direction override
                rule['dir'] = direction
            if not rule or rule.get('stub', False):
                if rule.get('stub', False):
                    del rule['stub']
                rule['target'] = target
                rule['dir'] = direction
                allrules[source] = rule

            # field is no longer a tree root
            if source in rules:
                del rules[source]

            target_rule = allrules.get(target, None)
            if target_rule is None:
                allrules[target] = target_rule = {
                    'stub': True,
                }
                rules[target] = target_rule
            if 'with' not in target_rule:
                target_rule['with'] = OrderedDict()
            rule['parent'] = target_rule
            target_rule['with'][source] = rule

        return rules

    def _process_field_moves(self, rules):
        """move fields according to the rules
        """
        for name, rule in rules.items():
            if name == '__all__':
                continue
            prefix = None
            if '.' in name:
                prefix, name = name.split('.', 1)
            else:
                prefix = ''
            if not rule.get('stub', False):
                after = rule['target'] if rule['dir'] == 'after' else None
                before = rule['target'] if rule['dir'] == 'before' else None
                if not (before or after):
                    raise ValueError(
                        'Direction of a field move must be before or '
                        'after, but got {0}.'.format(rule['direction'])
                    )
                try:
                    move(self, name, before=before, after=after, prefix=prefix)
                except KeyError as e:
                    if (
                        e.message.startswith('Field ') and
                        e.message.endswith(' not found')
                    ):
                        # The relative_to field doesn't exist
                        logger.warning(
                            'Field move to non-existing: '
                            'field name: {0}, rule: {1}'.format(
                                prefix + '.' + name,
                                str(rule)
                            )
                        )
                    else:
                        raise
            self._process_field_moves(rule.get('with', {}))

    def _process_group_order(self):
        self.groups.sort(key=attrgetter('order'))
